(** A few additions to Coquelicot.

    We add notations for Coquelicot operators in a new scope Rbar
    (i.e., R plus infinities). *)

Require Import Reals.
Require Import Coquelicot.Coquelicot.
Require Import nnR.

Declare Scope Rbar_scope.

Delimit Scope Rbar_scope with Rbar.

Bind Scope Rbar_scope with Rbar.

(** real : R -> Rbar being a coercion is a pain because it is not
    injective.

    Ideally we'd like to remove the coercion but this is not possible.
    We can at least ask Coq to always display it. *)
Add Printing Coercion real.

(** * Notation for Rbar less or equal *)

Infix "<=" := Rbar_le : Rbar_scope.

Arguments Rbar_le x y : simpl nomatch.

(** * Notation for sup *)

Reserved Notation "\sup_ ( i | P ) { F }"
  (format "'[' \sup_ ( i | P ) '/ ' { F } ']'").

Notation "\sup_ ( i | P ) { F }" :=
  (Rbar_lub (fun y => exists i, y = F /\ P)): Rbar_scope.

(** * Notations for non negative and non decreasing closures *)

Notation "''[' x ']+'" := (Rmax 0 x).

Notation "''[' f ']^'" := (fun t : nnR => \sup_(u | u <= t) { (f u)%R })%Rbar.

Notation "''[' f ']+^'" := ('[ fun t : nnR => '[ f t ]+ ]^).
