From mathcomp Require Import ssreflect ssrbool ssrnat ssrfun fintype finfun eqtype.
From mathcomp Require Import bigop ssralg.
Require Import Reals.
Require Import Coquelicot.Coquelicot.
Require Import mathcomp.analysis.Rstruct.
Require Import Psatz.

Set Implicit Arguments.
Unset Strict Implicit.
Unset Printing Implicit Defensive.

Delimit Scope R_scope with Re.
Delimit Scope ring_scope with Ri.

(** Add a sum notation in R_scope to avoid having to put scope anotations everytime *)

Notation "\sum_ i F" := (\sum_i (F : R)%Re)%Ri : R_scope.
Notation "\sum_ ( i < n ) F" := (\sum_(i < n) (F : R)%Re)%Ri : R_scope.
Notation "\sum_ ( i | P ) F" := (\sum_(i | P%nat) (F : R)%Re)%Ri : R_scope.
Notation "\sum_ ( i < n | P ) F" := (\sum_(i < n | P%nat) (F : R)%Re)%Ri : R_scope.

Lemma leq_sum I r (P : pred I) (E1 E2 : I -> R) :
  (forall i, P i -> (E1 i <= E2 i)%Re) ->
  ((\sum_(i <- r | P i) E1 i)%Ri <= (\sum_(i <- r | P i) E2 i)%Ri)%Re.
Proof.
by move=> leE12; (elim/big_ind2: _ ; try lra) => // m1 m2 n1 n2;
  unfold GRing.add; simpl; intros; lra.
Qed.

(** n-ary min and max on R *)

Reserved Notation "\max_ i F"
  (at level 41, F at level 41, i at level 0,
          right associativity,
          format "'[' \max_ i '/  '  F ']'").

Reserved Notation "\min_ i F"
  (at level 41, F at level 41, i at level 0,
          right associativity,
          format "'[' \min_ i '/  '  F ']'").

Notation "\min_ i F" :=
  (\big[Rmin/(([ffun i => F] ord0) : R)]_i (F : R)) : R_scope.

Notation "\max_ i F" :=
  (\big[Rmax/(([ffun i => F] ord0) : R)]_i (F : R)) : R_scope.

Lemma min_lb n i (F : R^n.+1) : \min_k F k <= F i.
Proof.
move: (_ ord0).
elim: n i F => [|n IH] i F x0.
{ rewrite big_ord_recl big_ord0.
  by apply (Rle_trans _ _ _ (Rmin_l _ _)); right; rewrite !zmodp.ord1. }
rewrite big_ord_recl.
case (@eqP _ ord0 i).
{ by move=> <-; apply (Rle_trans _ _ _ (Rmin_l _ _)); right. }
move=> /eqP /unlift_some [{}i -> _].
apply (Rle_trans _ _ _ (Rmin_r _ _)).
pose F' := [ffun i => F (lift ord0 i)].
rewrite (_ : F _ = F' i); [|by rewrite /F' ffunE].
under eq_bigr => j _ do [rewrite (_ : F _ = F' j); [|by rewrite /F' ffunE]].
apply IH.
Qed.

Lemma max_ub n i (F : R^n.+1) : F i <= \max_k F k.
Proof.
move: (_ ord0).
elim: n i F => [|n IH] i F x0.
{ rewrite big_ord_recl big_ord0.
  by refine (Rle_trans _ _ _ _ (Rmax_l _ _)); right; rewrite zmodp.ord1. }
rewrite big_ord_recl.
case (@eqP _ ord0 i).
{ by move=> <-; refine (Rle_trans _ _ _ _ (Rmax_l _ _)); right. }
move=> /eqP /unlift_some [{}i -> _].
refine (Rle_trans _ _ _ _ (Rmax_r _ _)).
pose F' := [ffun i => F (lift ord0 i)].
rewrite (_ : F _ = F' i); [|by rewrite /F' ffunE].
under eq_bigr => j _ do [rewrite (_ : F _ = F' j); [|by rewrite /F' ffunE]].
apply IH.
Qed.
