(** * Proof of Theorem 5 *)

Require Import mathcomp.ssreflect.ssreflect
        mathcomp.ssreflect.ssrbool mathcomp.ssreflect.ssrnat
        mathcomp.ssreflect.ssrfun mathcomp.ssreflect.eqtype
        mathcomp.ssreflect.seq mathcomp.ssreflect.path mathcomp.ssreflect.order
        mathcomp.ssreflect.fintype mathcomp.ssreflect.finfun
        mathcomp.ssreflect.bigop
        mathcomp.algebra.ssralg.
Require Import Flocq.Core.Raux.
Require Import Reals mathcomp.analysis.Rstruct nnR Coquelicot.Rbar Rbar Rbigop PWC.
Require Import Psatz.

Set Implicit Arguments.
Unset Strict Implicit.
Unset Printing Implicit Defensive.

Import Order.Theory.

Delimit Scope R_scope with Re.
Delimit Scope ring_scope with Ri.

Local Open Scope ring_scope.
Local Open Scope R_scope.

Section Theorem5.

(** ** Define all variables of the theorem *)

Variable n' : nat.
Notation n := n'.+1.  (** Be sure that n is non zero *)

Variable R L : R+*.

Variable tC tT : R+* ^ n.

Variable tJ : R+ ^ n.

Notation "'C`_' i" := (tC i) (at level 3, i at level 2).
Notation "'T`_' i" := (tT i) (at level 3, i at level 2).
Notation "'J`_' i" := (tJ i) (at level 3, i at level 2).

Hypothesis R_large_enough : \sum_i C`_i / T`_i < R.

(** *** A few definitions to make things more readable *)

Definition R' := R - \sum_i C`_i / T`_i.

Definition W := \sum_i (T`_i + J`_i - C`_i / R) * (C`_i / T`_i).

Definition Gl := (\min_k C`_k) * ((\sum_i C`_i / T`_i) - \max_i (C`_i / T`_i)).

Definition Gq :=
  \sum_(i < n) \sum_(j < n | j < i) Rmin T`_i T`_j * (C`_i * C`_j) / (T`_i * T`_j).

Definition C := R * L + W - Rmax Gl Gq / R.

Definition V t := \sum_i C`_i * IZR (Zceil ((t + J`_i) / T`_i)).

Definition beta R L := fun t : R+ => R * '[ t - L ]+.

(** ** Here starts the proof (see at bottom for the statement of the theorem itself) *)

(** *** First a few simple properties *)

Lemma V_non_decr : forall x y, x <= y -> V x <= V y.
Proof.
move=> x y Hxy; apply leq_sum => i _.
apply Rmult_le_compat_l; [lra'|].
apply IZR_le, Zceil_le, Rmult_le_compat_r; [|lra].
apply Rlt_le, Rinv_0_lt_compat; move: (posR_pos (T`_i)); lra.
Qed.

Lemma PCdT : forall k, 0 < C`_k / T`_k.
Proof.
move=> k.
apply (Rmult_lt_reg_r (T`_k)); [lra'|].
rewrite Rmult_0_l Rmult_assoc (Rinv_l _ (Rgt_not_eq _ _ (posR_pos (T`_k)))) Rmult_1_r.
move: (posR_pos (C`_k)); set c := C`_k; lra.
Qed.

Definition rho := \sum_i C`_i / T`_i.  (** just a convenient shortcut *)

Lemma Prho : 0 < rho.
Proof.
rewrite /rho.
rewrite big_ord_recl /= -(Rplus_0_r 0).
apply Rplus_lt_le_compat; [by apply PCdT|].
apply big_ind; [by apply Rle_refl|by rewrite /GRing.add /= => x y; lra|].
move=> i _; apply Rlt_le, PCdT.
Qed.

(** *** Most of the proof is parameterized by a real M *)

Section ForallM.

Variable M : R_ringType.  (** The type of real numbers [R] is masked by variable [R]. *)

(** *** Point 1 *)

(** { x | P x } is a Coq notation meaning that there exists some x such that P x. *)
Lemma point_1abc :
  { SM : R_ringType | V SM + M = R * (SM - L)
             /\ (forall t, t < SM -> R * (t - L) < V t + M)
             /\ (M + R * L + \sum_i J`_i * (C`_i / T`_i)) / R' <= SM }.
Proof.
have PinvsR : 0 < / R; [lra'|].
have PinvsRmrho : 0 < / (R - rho).
{ apply Rinv_0_lt_compat; rewrite /rho; lra. }
pose b := \sum_i J`_i * (C`_i / T`_i).
pose b' := b + \sum_i C`_i.
have Hbb' : b <= b'.
{ rewrite /b'.
  apply (Rplus_le_reg_r (- b)); ring_simplify.
  apply big_ind; [by apply Rle_refl|by rewrite /GRing.add /= => x y; lra|].
  move=> i _; apply nnR_pos. }
have Hbb't : forall t, rho * t + b <= V t <= rho * t + b'.
{ move=> t; rewrite /V /rho /b' /b; split.
  { set a := \sum__ _; set c := \sum__ _.
    rewrite -[a * t + c]/(a * t + c)%Ri /a /c big_distrl -big_split /=.
    apply leq_sum => i _.
    rewrite /GRing.add /GRing.mul /=.
    set ci := C`_i.
    rewrite (_ : _ + _ = ci * ((t + J`_i) / T`_i)) /ci;
      [|by rewrite -/ci; set ti := T`_i; set ji := J`_i; field; rewrite /ti; lra'].
    by apply Rmult_le_compat_l; [apply nnR_pos|apply Zceil_ub]. }
  set a := \sum__ _; set c := \sum__ _; set d := \sum__ _; set e := \sum__ _.
  rewrite -[c * t + (d + e)]/(c * t + (d + e))%Ri /a /c /d /e.
  rewrite big_distrl -!big_split /=.
  apply leq_sum => i _.
  rewrite /GRing.add /GRing.mul /=.
  set ci := C`_i; set ti := T`_i.
  rewrite (_ : _ + (_ + ci) = ci * ((t + J`_i) / ti + 1)) /ci /ti;
    [|by field_simplify; [|lra'..]].
  by apply Rmult_le_compat_l; [apply nnR_pos|apply Rlt_le, Zceil_lb]. }
pose x := (b + R * L + M) / (R - rho).
pose x' := (b' + R * L + M) / (R - rho).
have Hx : rho * x + b = R * (x - L) - M; [by rewrite /x /rho; field'|].
have Hx' : rho * x' + b' = R * (x' - L) - M; [by rewrite /x' /rho; field'|].
have Hxx' : x <= x'.
{ rewrite /x /x'.
  apply (Rmult_le_reg_r (R - rho)); [by rewrite /rho; lra|].
  rewrite !Rmult_assoc Rinv_l ?Rmult_1_r; [|rewrite /rho]; lra. }
have Hxx't : forall t, x <= t <= x' -> rho * x + b <= V t <= rho * x' + b'.
{ move=> t Ht.
  move: (Hbb't t).
  have H := Prho.
  have H' : rho * x <= rho * t; [by apply Rmult_le_compat_l; lra|].
  have H'' : rho * t <= rho * x'; [apply Rmult_le_compat_l|]; lra. }
have Hlx : forall t, t < x -> R * (t - L) - M < V t.
{ move=> t Ht.
  move: (proj1 (Hbb't t)); apply Rlt_le_trans.
  apply (Rplus_lt_reg_r (- rho * t + R * L + M)); ring_simplify.
  rewrite (_ : _ - _ = t * (R - rho)); [|ring].
  apply (Rmult_lt_reg_r (/ (R - rho))) => [//|].
  rewrite Rmult_assoc Rinv_r ?Rmult_1_r; [|by rewrite /rho; lra].
  apply (Rlt_le_trans _ _ _ Ht); right; rewrite /x; field; rewrite /rho; lra. }
case (RleP (V x) (R * (x - L) - M)) => [|/Rnot_le_lt] HVx.
{ exists x; split; [|split].
  { apply (Rplus_eq_reg_r (- M)).
    rewrite (_ : _ + (- M) = V x); [|ring].
    apply (Rle_antisym _ _ HVx).
    rewrite -Hx; apply Hbb't. }
  { move=> t Ht; move: (Hlx _ Ht); lra. }
  by rewrite -/b /R' -/rho Rplus_comm (Rplus_comm M) -Rplus_assoc; right. }
have {}Hlx : forall t, t <= x -> R * (t - L) - M < V t.
{ by move=> t [Ht|->]; [apply Hlx|]. }
have Hrx' : forall t, x' < t -> V t < R * (t - L) - M.
{ move=> t Ht.
  apply (Rle_lt_trans _ _ _ (proj2 (Hbb't t))).
  apply (Rplus_lt_reg_r (- rho * t + R * L + M)); ring_simplify.
  rewrite (_ : _ + t * R = t * (R - rho)); [|ring].
  apply (Rmult_lt_reg_r (/ (R - rho))) => [//|].
  rewrite Rmult_assoc Rinv_r ?Rmult_1_r; [|by rewrite /rho; lra].
  move: Ht; apply Rle_lt_trans; right; rewrite /x'; field; rewrite /rho; lra. }
have : { p : PWC |
         in_segment p x x' /\ forall t, x <= t <= x' -> V t = PWC2f p t }.
{ have H := [ffun i => nuPWC (T`_i) (C`_i) (J`_i) x x'].
  exists (\big[PWC_add/PWC_cst 0]_i proj1_sig (H i)); split.
  { apply big_ind => [//| |].
    { by move=> f g Hf Hg; apply PWC_add_segment. }
    by move=> i _ /=; move: (H i) => [f [H' _]] /=. }
  move=> t Ht; rewrite /V.
  under eq_bigr => i _ do move: (proj2_sig (H i)) => [_ /(_ _ Ht) <-].
  symmetry; apply (big_morph (fun f => PWC2f f t)) => [|//].
  move=> f g; apply PWC_add_correct. }
move=> [p] [Hpx Hp].
suff : { SM : R_ringType |
         x < SM <= x' /\ PWC2f p SM = R * (SM - L) - M
         /\ forall t, t < SM -> R * (t - L) - M < PWC2f p t }.
{ move=> [SM [HSM [H1 H2]]]; exists SM; split; [|split].
  { rewrite (Hp _ (conj (Rlt_le _ _ (proj1 HSM)) (proj2 HSM))) H1; ring. }
  move=> t HtSM.
  apply (Rplus_lt_reg_r (- M)); rewrite (_ : V t + M + - M = V t); [|ring].
  case (RleP x t) => [|/Rnot_le_lt] Htx.
  { rewrite (Hp _ (conj Htx (Rlt_le _ _ (Rlt_le_trans _ _ _ HtSM (proj2 HSM))))).
    by apply H2. }
  { by apply Hlx, Rlt_le. }
  move: (Rlt_le _ _ (proj1 HSM)); apply Rle_trans.
  by rewrite -/b /R' -/rho Rplus_comm (Rplus_comm M) -Rplus_assoc; right. }
have {}Hlx : forall t, t <= x -> R * (t - L) - M < PWC2f p t.
{ move=> t Ht.
  apply (Rlt_le_trans _ _ _ (Hlx _ Ht)).
  rewrite (PWC_left_segment Hpx Ht).
  apply (Rle_trans _ _ _ (V_non_decr Ht)).
  right; apply Hp; lra. }
have {}Hrx' : forall t, x' < t -> PWC2f p t < R * (t - L) - M.
{ move=> t Ht.
  move: (Hrx' _ Ht); apply Rle_lt_trans.
  rewrite (PWC_right_segment Hpx (Rlt_le _ _ Ht)).
  move: (V_non_decr (Rlt_le _ _ Ht)); apply Rle_trans.
  right; symmetry; apply Hp; lra. }
move: x Hxx' Hpx Hlx Hrx' {Hx Hxx't Hp HVx}.
case p => s c Hs1 Hs2.
elim: s c Hs1 Hs2 => [|[hs1 hs2] ts IHs] c Hs1 Hs2 x Hxx' Hpx Hlx Hrx'.
{ pose SM := (c + M) / R + L.
  exists SM; split; split.
  { apply Rnot_le_lt => HSMx.
    move: (Hlx _ HSMx) => H.
    move: (Rlt_le_trans _ _ _ H (PWC_non_decr _ HSMx)).
    rewrite /PWC2f /= /SM (_ : _ - M = c); [apply Rlt_irrefl|field']. }
  { apply Rnot_lt_le => Hx'SM.
    move: (Hrx' _ Hx'SM) => H.
    move: (Rle_lt_trans _ _ _ (PWC_non_decr _ (Rlt_le _ _ Hx'SM)) H).
    rewrite /PWC2f /= /SM (_ : _ - M = c); [apply Rlt_irrefl|field']. }
  { rewrite /PWC2f /= /SM; field'. }
  move=> t Ht.
  case (RleP t x) => [|/Rnot_le_lt] Hxt; [by apply Hlx|].
  apply (Rplus_lt_reg_r M), (Rmult_lt_reg_r (/ R)) => [//|].
  apply (Rplus_lt_reg_r L).
  rewrite (_ : _ + L = t); [apply Ht|field']. }
case (RleP hs2 (R * (hs1 - L) - M)) => [|/Rnot_le_lt] Hhs2i.
{ pose SM := (hs2 + M) / R + L.
  have HSMhs1 : Rleb SM hs1.
  { apply /RleP.
    rewrite /SM.
    apply (Rplus_le_reg_r (- L)).
    apply (Rmult_le_reg_r R); [lra'|].
    rewrite (_ : _ * R = hs2 + M); [|field]; lra'. }
  exists SM; split; split.
  { apply Rnot_le_lt => HSMx.
    move: (Hlx _ HSMx) => H.
    move: (Rlt_le_trans _ _ _ H (PWC_non_decr _ HSMx)).
    move: Hpx; rewrite /in_segment /PWC2f /= => /andP [] /andP [] -> _ _ /=.
    rewrite /SM (_ : _ - M = hs2); [apply Rlt_irrefl|field']. }
  { apply Rnot_lt_le => Hx'SM.
    move: (Hrx' _ Hx'SM) => H.
    move: (Rle_lt_trans _ _ _ (PWC_non_decr _ (Rlt_le _ _ Hx'SM)) H).
    rewrite (_ : _ - M = hs2) /SM; [|field'].
    apply Rle_not_lt.
    apply (Rle_trans _ (PWC2f (Build_PWC Hs1 Hs2) hs1)).
    { right.
      by rewrite /PWC2f /= (_ : Rleb hs1 hs1 = true) //; apply /RleP; right. }
    apply PWC_non_decr.
    move: Hpx; rewrite /in_segment /= => /andP [] /andP [] _ /RltP; lra. }
  { rewrite /PWC2f /= HSMhs1 /= /SM; field'. }
  move=> t Ht.
  case (RleP t x) => [|/Rnot_le_lt] Hxt; [by apply Hlx|].
  rewrite /PWC2f /=.
  have -> /= : Rleb t hs1; [by apply /RleP; move: HSMhs1 => /RleP; lra|].
  apply (Rplus_lt_reg_r M), (Rmult_lt_reg_r (/ R)) => [//|].
  apply (Rplus_lt_reg_r L).
  rewrite (_ : _ + L = t); [apply Ht|field']. }
have Hs'1 : sorted Rleb (unzip1 ts).
{ move: (Hs1) => /=; apply path_sorted. }
have Hs'2 : sorted Rleb (rcons (unzip2 ts) c).
{ move: (Hs2) => /=; apply path_sorted. }
have Hs1x' : hs1 <= x'.
{ move: Hpx; rewrite /in_segment /= => /andP [] /andP [] _ /RltP; lra. }
have Hpx' : in_segment (Build_PWC Hs'1 Hs'2) hs1 x'.
{ apply /allP => t /= Ht.
  move: Hpx => /allP /= /(_ t).
  rewrite in_cons orbC Ht /= => /(_ (eq_refl true)) /andP [] /RleP Hxt ->.
  rewrite andbC /=; apply /RleP.
  have H : transitive Rleb; [exact: le_trans|].
  by move: (Hs1) => /= /(order_path_min H) {H} /allP /(_ _ Ht) /RleP. }
have Hl'x : forall t, t <= hs1 -> R * (t - L) - M < PWC2f (Build_PWC Hs'1 Hs'2) t.
{ move=> t Ht.
  apply (Rlt_le_trans _ hs2).
  { move: Hhs2i; apply Rle_lt_trans, Rplus_le_compat_r.
    by apply Rmult_le_compat_l; lra'. }
  rewrite /PWC2f /=.
  set img := nth _ _ _.
  have H : transitive Rleb; [exact: le_trans|].
  apply /RlebP.
  move: (Hs2) => /= /(order_path_min H) {H} /allP /(_ img); apply.
  rewrite mem_rcons in_cons; apply /orP.
  case (leqP (size (unzip2 ts)) (find (Rleb t) (unzip1 ts))) => Hs.
  { by left; apply /eqP; rewrite /img nth_default. }
  by right; apply mem_nth. }
have Hr'x' : forall t, x' < t -> PWC2f (Build_PWC Hs'1 Hs'2) t < R * (t - L) - M.
{ move=> t Ht; move: (Hrx' _ Ht); apply Rle_lt_trans.
  rewrite /PWC2f /=.
  have -> : Rleb t hs1 = false; [by apply /RleP /Rlt_not_le; lra|].
  by rewrite /=; right. }
move: (IHs c Hs'1 Hs'2 hs1 Hs1x' Hpx' Hl'x Hr'x') => [SM [HSM1 [HSM2 HSM3]]].
exists SM.
split; split.
{ move: (proj1 HSM1); apply Rle_lt_trans.
  by move: Hpx; rewrite /in_segment /= => /andP [] /andP [] /RleP. }
{ exact (proj2 HSM1). }
{ rewrite -HSM2 /PWC2f /=.
  have -> //= : Rleb SM hs1 = false.
  apply /RleP /Rlt_not_le; lra. }
move=> t Ht.
rewrite /PWC2f /=.
case RleP => [|/Rnot_le_lt] Hths1 /=.
{ move: Hhs2i; apply Rle_lt_trans; move: (nnR_pos R); nra. }
by apply (Rlt_le_trans _ _ _ (HSM3 _ Ht)); right.
Qed.

(** The SM whose existence we just proved. *)
Definition SM : R_ringType := proj1_sig point_1abc.

Lemma point_1a : V SM + M = R * (SM - L).
Proof. by rewrite /SM; case: point_1abc => SM [H _] /=. Qed.

Lemma point_1b : forall t, t < SM -> R * (t - L) - M < V t.
Proof. rewrite /SM; case: point_1abc => SM [_ [H _]] /= t; move: (H t); lra. Qed.

Lemma point_1c : (M + R * L + \sum_i J`_i * (C`_i / T`_i)) / R' <= SM.
Proof. by rewrite /SM; case: point_1abc => SM [_ [_ H]] /=. Qed.

(** *** Point 2 *)

Definition qM : R_ringType^n := [ffun i => IZR (Zceil ((SM + J`_i) / T`_i))].

Definition rM : R_ringType^n := [ffun i => T`_i * qM i - (SM + J`_i)].

Lemma point_2a : forall i, qM i = (SM + J`_i + rM i) / T`_i.
Proof. move=> i; rewrite /rM !ffunE /=; set qMi := IZR _; field'. Qed.

Lemma nnrM : forall i, 0 <= rM i.
Proof.
move=> i.
rewrite /rM /qM !ffunE.
apply (Rplus_le_reg_r (SM + J`_i)); rewrite Rplus_0_l.
apply (Rmult_le_reg_r (/ T`_i));
  [apply Rinv_0_lt_compat; move: (posR_pos (T`_i)); lra|].
rewrite (_ : (_ + (_ + _)) * / T`_i = IZR (Zceil ((SM + J`_i) / T`_i))); [|field'].
apply Zceil_ub.
Qed.

Lemma point_2b : forall i, rM i < T`_i.
Proof.
move=> i.
rewrite /rM /qM !ffunE.
apply (Rplus_lt_reg_r (SM + J`_i)), (Rmult_lt_reg_r (/ T`_i)); [lra'|].
rewrite (_ : (_ + (_ + _)) * / T`_i = IZR (Zceil ((SM + J`_i) / T`_i))); [|field'].
rewrite (_ : _ * / T`_i = (SM + J`_i) / T`_i + 1); [|field'].
apply Zceil_lb.
Qed.

(** *** Point 3 *)

Lemma point_3 :
  R' * SM = M + R * L + (\sum_i C`_i / T`_i * J`_i) + \sum_i C`_i / T`_i * rM i.
Proof.
rewrite /R' Rplus_assoc.
rewrite -big_split /=.
under [in RHS] eq_bigr => i _ do rewrite /GRing.add /= -Rmult_plus_distr_l.
apply (Rplus_eq_reg_r (- R * L + SM * \sum_i C`_i / T`_i)).
rewrite (_ : _ + _ = R * (SM - L)); [|by set s := \sum__ _; ring].
rewrite [RHS](_ : _ = (SM * \sum_i (C`_i / T`_i)) + \sum_i (C`_i / T`_i * (J`_i + rM i)) + M);
  [|by rewrite /GRing.add /=; set s := \sum__ _; set s' := \sum__ _; ring].
rewrite big_distrr -big_split /=.
under eq_bigr => i _ do rewrite Rmult_comm -Rmult_plus_distr_l -Rplus_assoc.
under eq_bigr => i _ do rewrite Rmult_assoc (Rmult_comm (/ T`_i)).
under eq_bigr => i _ do [move: point_2a; rewrite /Rdiv => <-; rewrite /qM !ffunE].
by rewrite -/(V SM) point_1a.
Qed.

(** *** Point 4 *)

Definition lM : R_ringType^n := [ffun i => (qM i - 1) * T`_i - J`_i].

Lemma point_4a : forall i, lM i < SM.
Proof.
move=> i.
rewrite /lM ffunE.
apply (Rplus_lt_reg_r (-SM + T`_i)); rewrite (_ : SM + _ = T`_i); [|ring].
rewrite (_ : _ + _ = T`_i * qM i - (SM + J`_i)); [|ring].
rewrite (_ : _ - _ = rM i); [|by rewrite /rM !ffunE].
apply point_2b.
Qed.

(** ['I_n] is the set { 0,..., n-1 } of natural numbers less than [n]

So sigma : ['I_n^n] is a functions from ['I_n] to itself. *)
Lemma sigma_proof :
  { sigma : 'I_n^n |
    injective sigma
    /\ forall i j : 'I_n, (i <= j)%nat -> lM (sigma j) <= lM (sigma i) }.
Proof.
move: (perm_iota_sort Rgeb 0 [seq lM i | i <- ord_enum n]).
move=> [sigma_s].
rewrite size_map -(size_map val) val_ord_enum size_iota.
move=> Hperm.
have Ssigma : size sigma_s = n; [by rewrite (perm_size Hperm) size_iota|].
rewrite (proj1 (eq_in_map (nth 0 _) (fun i => lM (inord i)) _)); last first.
{ move=> i Hi.
  set f := fun i => lM (inord i); rewrite -/(f i).
  under eq_map => x do [rewrite (_ : lM x = f x); [|by rewrite /f inord_val]].
  rewrite /f 2!map_comp val_ord_enum -map_comp.
  rewrite -[map _ _]/(mkseq f n) nth_mkseq //.
  by move: Hi; rewrite (perm_mem Hperm) mem_iota. }
move=> Hsort.
have {}Hsort : sorted Rgeb [seq lM (inord i) | i <- sigma_s].
{ rewrite -Hsort; apply sort_sorted.
  have : total Rleb; [exact: le_total|].
  by move=> H x y; move: (H x y) => /orP [|] {}H; apply /orP; [right|left]. }
exists [ffun i : 'I_n => inord (nth O sigma_s i)]; split.
{ move=> i j; rewrite !ffunE.
  have Hi : (i < size sigma_s)%nat; [by rewrite Ssigma; case i|].
  have Hj : (j < size sigma_s)%nat; [by rewrite Ssigma; case j|].
  have Hsigma : all (gtn n) sigma_s.
  { by rewrite (perm_all _ Hperm); apply /allP => k; rewrite mem_iota. }
  move=> /(congr1 val) /=.
  rewrite !inordK => [H| |]; [|by move: Hsigma => /all_nthP; apply..].
  move: (perm_uniq Hperm); rewrite iota_uniq => /uniqP H'.
  by apply ord_inj, (H' 0%nat). }
move=> i j Hij.
rewrite (ffunE _ i) (ffunE _ j).
apply /RlebP.
move: Hsort.
rewrite sorted_map.
set r := relpre _ _.
move=> Hsort.
rewrite -[Rleb _ _]/(r (nth O sigma_s i) (nth O sigma_s j)).
have Htrans : transitive r.
{ by move=> x y z /RleP Hxy /RleP Hyz; apply /RleP; move: Hxy; apply Rle_trans. }
have Hrefl : reflexive r.
{ by move=> x; apply /RleP; right. }
apply (sorted_leq_nth Htrans Hrefl O Hsort).
{ by rewrite Ssigma; case i. }
{ by rewrite Ssigma; case j. }
exact Hij.
Qed.

Definition sigma : 'I_n^n := proj1_sig sigma_proof.

(** [sigma] is an injection from ['I_n] to itself, meaning it is a permutation of ['I_n] *)
Lemma injective_sigma : injective sigma.
Proof. by rewrite /sigma; case sigma_proof => /= s []. Qed.

(** and it sorts [lM] by decreasing order *)
Lemma sigma_order :
  forall i j : 'I_n, (i <= j)%nat -> lM (sigma j) <= lM (sigma i).
Proof. by rewrite /sigma; case sigma_proof => /= s []. Qed.

(** *** Point 5 *)

Lemma point_5 : forall k,
  C`_k * IZR (Zceil ((SM + J`_k) / T`_k))
  = C`_k * IZR (Zceil ((lM k + J`_k) / T`_k)) + C`_k.
Proof.
move=> k.
rewrite -{3}(Rmult_1_r (C`_k)) -Rmult_plus_distr_l.
apply f_equal2 => [//|].
rewrite (_ : IZR _ = qM k); [|by rewrite /qM ffunE].
apply (Rplus_eq_reg_r (- 1)); ring_simplify.
rewrite /lM !ffunE (_ : (_ - _ + _) / T`_k = qM k - 1) /qM ffunE; [|field'].
by rewrite -minus_IZR Zceil_IZR.
Qed.

(** *** Point 6 *)

Lemma point_6 : forall i,
  lM (sigma i) + \sum_(k < n | k <= i) C`_(sigma k) / R <= SM.
Proof.
move=> i.
pose Si := [seq sigma k | k : 'I_n <- ord_enum n & (k <= i)%nat].
have HSi : forall k, k \in Si ->
  C`_k * IZR (Zceil ((lM (sigma i) + J`_k) / T`_k)) + C`_k
  <= C`_k * IZR (Zceil ((SM + J`_k) / T`_k)).
{ move=> k Hk; rewrite point_5.
  apply Rplus_le_compat_r, Rmult_le_compat_l; [lra'|].
  apply IZR_le, Zceil_le, Rmult_le_compat_r; [lra'|].
  apply Rplus_le_compat_r.
  move: Hk => /mapP [k']; rewrite mem_filter => /andP [] Hk' _ ->.
  by apply sigma_order. }
have HSti : forall k, k \notin Si ->
  C`_k * IZR (Zceil ((lM (sigma i) + J`_k) / T`_k))
  <= C`_k * IZR (Zceil ((SM + J`_k) / T`_k)).
{ move=> k Hk.
  apply Rmult_le_compat_l; [lra'|].
  apply IZR_le, Zceil_le, Rmult_le_compat_r; [lra'|].
  apply Rplus_le_compat_r, Rlt_le, point_4a. }
apply (Rmult_le_reg_l R); [lra'|].
apply (Rplus_le_reg_r (- R * L - M)).
rewrite (_ : R * SM + _ = R * (SM - L) - M); [|ring].
rewrite -point_1a (_ : V SM + M - M = V SM); [|ring].
rewrite (_ : _ + _ = R * (lM (sigma i) - L) - M + ((\sum_(k < n | (k <= i)%nat) (C`_(sigma k) / R)%Re) * R)%Ri);
  [|rewrite /GRing.mul /=; ring].
rewrite big_distrl /=.
under eq_bigr => k _ do [rewrite /GRing.mul /= Rmult_assoc Rinv_l ?Rmult_1_r; [|lra']].
apply (Rle_trans _ (V (lM (sigma i)) + \sum_(k < n | (k <= i)%nat) C`_(sigma k))).
{ apply Rplus_le_compat_r, Rlt_le, point_1b, point_4a. }
have Hl : forall k : 'I_n, (k <= i)%nat = (sigma k \in Si).
{ move=> k.
  apply/idP/idP => [Hk|]; rewrite /Si (mem_map injective_sigma) mem_filter.
  { rewrite Hk /=; apply mem_ord_enum. }
  by move=> /andP []. }
under eq_bigl => k do rewrite Hl.
rewrite -(reindex_inj (P:=fun k => k \in Si) (F:=fun k => C`_k : R_ringType)
                      injective_sigma) /=.
rewrite /V.
rewrite (bigID (fun i => i \in Si)) /= /GRing.add /= Rplus_comm -Rplus_assoc.
rewrite -big_split /=.
rewrite [in X in _ <= X](bigID (fun i => i \in Si)) /=.
apply Rplus_le_compat.
{ by apply leq_sum => k Hk; rewrite Rplus_comm; apply HSi. }
by apply leq_sum => k Hk; apply HSti.
Qed.

(** *** Point 7 *)

Lemma point_7 : forall i,
  rM (sigma i) <= T`_(sigma i) - \sum_(j < n | j <= i) C`_(sigma j) / R.
Proof.
move=> i.
rewrite /rM ffunE.
rewrite -(Ropp_involutive (_ -_)) -[in X in _ <= X](Ropp_involutive (_ -_)).
apply Ropp_le_contravar.
apply (Rplus_le_reg_r (T`_(sigma i) * qM (sigma i) - J`_(sigma i))).
rewrite [in X in _ <= X](_ : _ + _ = SM); [|ring].
rewrite (_ : _ + _ = (qM (sigma i) - 1) * T`_(sigma i) - J`_(sigma i)
                     + \sum_(j < n | (j <= i)%nat) (C`_(sigma j) / R)%Re); [|ring].
rewrite (_ : _ - _ = lM (sigma i)); [|by rewrite /lM !ffunE].
apply point_6.
Qed.

(** *** Point 8 *)

Lemma point_8 :
  \sum_i rM i * (C`_i / T`_i)
  <= \sum_i (T`_i - C`_i / R) * (C`_i / T`_i)
     - 1 / R * \sum_(i < n) \sum_(j < n | j < i)
                (C`_(sigma i) * C`_(sigma j)) / T`_(sigma i).
Proof.
rewrite (reindex_inj injective_sigma) /=.
rewrite [X in _ <= X - _](reindex_inj injective_sigma) /=.
under [in X in _ <= _ - X] eq_bigr => i _ do
  ((under eq_bigr => j _ do rewrite Rmult_comm /Rdiv Rmult_assoc);
   rewrite -big_distrl /=).
rewrite big_distrr /=.
under [in X in _ <= _ - X] eq_bigr => i _ do rewrite -Rmult_assoc big_distrr /=.
rewrite (_ : ?[a] - ?[b] = (?a - ?b)%Ri); [|by []].
rewrite -GRing.sumrB.
apply leq_sum=> i _.
rewrite /GRing.add /GRing.opp /=.
rewrite Ropp_mult_distr_l /Rdiv -Rmult_plus_distr_r.
apply Rmult_le_compat_r.
{ apply Rmult_le_pos; [lra'|].
  apply Rlt_le, Rinv_0_lt_compat.
  move: (posR_pos (T`_(sigma i))); set t := tT _; lra. }
apply (Rle_trans _ _ _ (point_7 i)).
rewrite /Rminus Rplus_assoc; apply Rplus_le_compat_l.
rewrite -Ropp_plus_distr; apply Ropp_le_contravar; right; symmetry.
rewrite (bigD1 i) => [/=|//].
rewrite /Rdiv; apply Rplus_eq_compat_l, eq_big.
{ by move=> j; rewrite ltn_neqAle andbC. }
move=> j _; field'.
Qed.

(** *** Point 9 *)

Lemma point_9a :
  Gl <= \sum_(i < n) \sum_(j < n | j < i) (C`_(sigma j) * C`_(sigma i)) / T`_(sigma i).
Proof.
rewrite /Gl.
apply (Rle_trans _ ((\min_k C`_k) * \sum_i C`_(sigma i) / T`_(sigma i) * INR i)).
{ apply Rmult_le_compat_l.
  { by apply big_ind => [|x y Hx Hy|i _];
      [rewrite ffunE; apply nnR_pos|apply Rmin_case|apply nnR_pos]. }
  rewrite [in X in _ <= X](bigID (fun i : 'I_n => i == O :> nat)) /=.
  under [in X in _ <= X + _] eq_bigr => i /eqP -> do rewrite Rmult_0_r.
  rewrite GRing.sumr_const /= GRing.mul0rn Rplus_0_l.
  set m := \max__ _.
  apply (Rplus_le_reg_r m); rewrite (_ : ?[a] - m + m = ?a); [|ring].
  rewrite (reindex_inj injective_sigma) /=.
  rewrite (bigID (fun i : 'I_n => i == O :> nat)) /= Rplus_comm.
  apply Rplus_le_compat.
  { rewrite (big_pred1 ord0) => [|//].
    rewrite (_ : m = \max_k ([ffun k => C`_k / T`_k] k));
      [|by rewrite ffunE; under eq_bigr => k _ do rewrite ffunE].
    by refine (Rle_trans _ _ _ _ (max_ub (sigma ord0) _)); right; rewrite ffunE. }
  apply leq_sum => [] [i Hi] /= /eqP Hi'.
  case: i Hi Hi' => [//|i] Hi _.
  rewrite -{1}(Rmult_1_r (_ / _)); apply Rmult_le_compat_l; [by apply Rlt_le, PCdT|].
  rewrite S_INR Rplus_comm -{1}(Rplus_0_r 1); apply Rplus_le_compat_l, pos_INR. }
rewrite big_distrr /=.
apply leq_sum => i _.
rewrite (_ : _ * _ = C`_(sigma i) / T`_(sigma i) * ((\min_k C`_k) * INR i));
  [|by set m := \big[_/_]__ _; set d := _ / _; ring].
under [in X in _ <= X] eq_bigr => j _ do rewrite /Rdiv Rmult_assoc Rmult_comm.
rewrite -big_distrr /= /GRing.mul /= /Rdiv.
apply Rmult_le_compat_l; [by apply Rlt_le, PCdT|].
rewrite (_ : _ * _ = \sum_(k < n | (k < i)%nat) \min_k C`_k).
{ apply leq_sum => j _.
  rewrite (_ : \big[_/_]__ _ = \min_k [ffun k => C`_k : R_ringType] k);
    [|by symmetry; rewrite !ffunE; under eq_bigr => k _ do rewrite ffunE].
  rewrite (_ : C`_(sigma j) = [ffun k => C`_k : R_ringType] (sigma j) :> R_ringType); [|by rewrite ffunE].
  apply min_lb. }
under [RHS] eq_bigr => j _ do rewrite -(Rmult_1_r (\big[_/_]__ _)).
rewrite -big_distrr /= /GRing.mul /=.
apply Rmult_eq_compat_l.
rewrite -(big_mkord (fun j => (j < i)%nat) (fun _ => 1)).
under eq_bigl => j do rewrite -[(j < i)%nat]/(true && (j < i)%nat).
rewrite -(big_nat_widen O i n); [|by apply ltnW; case i].
rewrite big_const_nat subn0.
case i => {}i _ /=; elim: i => [//|i IH].
by rewrite S_INR /= -IH Rplus_comm.
Qed.

Lemma point_9b :
  \sum_(i < n) \sum_(j < n | j < i)
    (C`_(sigma i) * C`_(sigma j)) / (T`_(sigma i) * T`_(sigma j))
    * Rmin (T`_(sigma i)) (T`_(sigma j))
  <= \sum_(i < n) \sum_(j < n | j < i) (C`_(sigma i) * C`_(sigma j)) / T`_(sigma i).
Proof.
apply leq_sum => i _; apply leq_sum => j _.
rewrite /Rdiv !Rmult_assoc.
do 2 (apply Rmult_le_compat_l; [lra'|]).
apply (Rmult_le_reg_r (T`_(sigma j))); [lra'|].
rewrite (_ : _ * _ = / (T`_(sigma i)) * (Rmin (T`_(sigma i)) (T`_(sigma j)))); [|field'].
apply Rmult_le_compat_l; [lra'|apply Rmin_r].
Qed.

Lemma point_9c :
  \sum_(i < n) \sum_(j < n | j < i)
    (C`_(sigma i) * C`_(sigma j)) / (T`_(sigma i) * T`_(sigma j))
    * Rmin (T`_(sigma i)) (T`_(sigma j))
  = Gq.
Proof.
rewrite /Gq !pair_big_dep /=.
under eq_bigr => p _ do rewrite Rmult_comm.
set x := fun p q => Rmin (T`_p) (T`_q) * (C`_p * C`_q) / (T`_p * T`_q).
under eq_bigr => p _ do
  [rewrite (_ : _ * _ = x (sigma p.1) (sigma p.2)); [|by rewrite /x /Rdiv !Rmult_assoc]].
under [RHS] eq_bigr => p _ do rewrite -/(x p.1 p.2).
have Hx : forall i j, x i j = x j i; [by move=> i j; rewrite /x Rmin_comm; field'|].
pose h :=
  fun p : 'I_n * 'I_n =>
    if (p.2 < p.1)%nat == (sigma p.2 < sigma p.1)%nat then (sigma p.1, sigma p.2)
    else (sigma p.2, sigma p.1).
rewrite (eq_bigl (fun p => ((h p).2 < (h p).1)%nat)); last first.
{ move=> p; rewrite /h /=.
  case ltnP => Hp; case (ltnP (sigma p.2)) => /=.
  - by move=> ->.
  - rewrite leq_eqVlt => /orP [/eqP /ord_inj /injective_sigma H|-> //].
    by exfalso; move: Hp; rewrite H ltnn.
  - by rewrite ltnNge => /negP H; symmetry; apply /negP => H'; apply H, ltnW.
  - by rewrite leqNgt => /negP H; symmetry; apply /negP. }
pose F := fun p => x p.1 p.2.
under eq_bigr => p _ do
  [rewrite (_ : x _ _ = F (h p)); [|by rewrite /F /h; case eqP => _ /=; [|apply Hx]]].
symmetry; apply (reindex_inj (h:=h)).
move=> p q Hpq; rewrite (surjective_pairing p) (surjective_pairing q).
move: Hpq; rewrite /h.
case ltnP => H; case ltnP => H'; case ltnP => H''; case ltnP => H'''
  /= [] /injective_sigma H1 /injective_sigma H2;
  (try by rewrite H1 H2);
  (try by move: H H' H'' H'''; rewrite H1 H2;
   (move=> /ssrnat.ltP H || move=> /ssrnat.leP H);
   (move=> /ssrnat.ltP H' || move=> /ssrnat.leP H');
   (move=> /ssrnat.ltP H'' || move=> /ssrnat.leP H'');
   (move=> /ssrnat.ltP H''' || move=> /ssrnat.leP H''');
   exfalso; lia);
  (try by move: H' H'''; rewrite H1 H2 => H' H''';
   move: (andb_true_intro (conj H' H''')) => /anti_leq /ord_inj /injective_sigma ->);
  by move: H H''; rewrite H1 H2 => H H'';
  move: (andb_true_intro (conj H H'')) => /anti_leq /ord_inj ->.
Qed.

(** *** Point 10 *)

Lemma point_10 : SM <= (M + C) / R'.
Proof.
apply (Rmult_le_reg_r R'); [by rewrite /R'; lra|].
rewrite /Rdiv Rmult_assoc Rinv_l ?Rmult_1_r; [|by rewrite /R'; lra].
rewrite Rmult_comm point_3.
rewrite /C /Rminus !Rplus_assoc.
do 2 apply Rplus_le_compat_l.
rewrite /W.
under [in X in _ + X <= _] eq_bigr => i _ do rewrite Rmult_comm.
apply (Rle_trans _ _ _ (Rplus_le_compat_l _ _ _ point_8)).
rewrite /Rminus -Rplus_assoc -big_split /=.
apply Rplus_le_compat.
{ right; apply eq_bigr => i _; ring. }
apply Ropp_le_contravar, Rmax_case.
{ apply (Rmult_le_reg_l R); [lra'|].
  rewrite /Rdiv Rmult_1_l Rmult_comm Rmult_assoc Rinv_l ?Rmult_1_r; [|lra'].
  rewrite -Rmult_assoc Rinv_r ?Rmult_1_l; [|lra'].
  apply (Rle_trans _ _ _ point_9a).
  right; apply eq_bigr => i _; apply eq_bigr => j _.
  by rewrite Rmult_comm /Rdiv. }
apply (Rmult_le_reg_l R); [lra'|].
rewrite /Rdiv Rmult_1_l Rmult_comm Rmult_assoc Rinv_l ?Rmult_1_r; [|lra'].
rewrite -Rmult_assoc Rinv_r ?Rmult_1_l; [|lra'].
rewrite -point_9c.
apply (Rle_trans _ _ _ point_9b).
right; apply eq_bigr => i _; apply eq_bigr => j _.
by rewrite Rmult_comm /Rdiv.
Qed.

End ForallM.

(** ** The theorem itself *)

Theorem theorem5 :
  forall t, (beta R' (C / R') t <= '[ fun t => beta R L t - V t ]+^ t)%Rbar.
Proof.
move=> t.
rewrite /beta.
rewrite Rmult_comm Rcomplements.Rmax_mult ?Rmult_0_l; [|by rewrite /R'; lra].
set s := (\sup_(_|_){ _ })%Rbar.
have : (0 <= s)%Rbar.
{ apply (Rbar_le_trans _ '[ R * '[ 0 - L ]+ - V 0 ]+).
  { apply Rmax_l. }
  rewrite /s /Lub.Rbar_lub; case Lub.Rbar_ex_lub => l [Hl _] /=.
  apply Hl; exists (mk_nnR 0 (Rle_refl _)); split => [//|]; apply nnR_pos. }
case_eq s => [r Hs Pr /=|//|//].
apply Rmax_case => [//|].
rewrite (_ : _ * _ = R' * t - C); [|by field; rewrite /R'; lra].
case (RleP t (C / R')) => [|/Rnot_le_lt] Ht.
{ apply (Rplus_le_reg_r C), (Rmult_le_reg_r (/ R')).
  { apply Rinv_0_lt_compat; rewrite /R'; lra. }
  rewrite (_ : _ * / R' = t); [|by field; rewrite /R'; lra].
  have : 0 <= r * / R'.
  { apply Rmult_le_pos => [//|].
    apply Rlt_le, Rinv_0_lt_compat; rewrite /R'; lra. }
  lra. }
set M := R' * t - C.
have -> : M = R * (SM M - L) - V (SM M); [by rewrite -point_1a; ring|].
apply (Rle_trans _ '[R * '[SM M - L]+ - V (SM M)]+).
{ refine (Rle_trans _ _ _ _ (Rmax_r _ _)).
  by apply Rplus_le_compat_r, Rmult_le_compat_l; [apply nnR_pos|apply Rmax_r]. }
rewrite -[?[a] <= ?[b]]/(Finite ?[a] <= Finite ?[b])%Rbar -Hs.
rewrite /s /Lub.Rbar_lub; case Lub.Rbar_ex_lub => l [Hl _] /=.
apply Hl.
have PSM : 0 <= SM M.
{ move: (point_1c M); apply Rle_trans.
  apply (Rmult_le_reg_r R'); [by rewrite /R'; lra|].
  rewrite Rmult_0_l Rmult_assoc Rinv_l ?Rmult_1_r; [|by rewrite /R'; lra].
  apply Rplus_le_le_0_compat; [apply Rplus_le_le_0_compat|].
  { rewrite /M.
    apply (Rplus_le_reg_r C), (Rmult_le_reg_r (/ R')).
    { apply Rinv_0_lt_compat; rewrite /R'; lra. }
    rewrite [in X in _ <= X](_ : _ * _ = t); [|by field; rewrite /R'; lra].
    move: Ht; lra. }
  { apply Rmult_le_pos; apply nnR_pos. }
  apply big_ind => [|x y Hx Hy|i _]; [by right|by rewrite /GRing.add /=; lra|].
  apply Rmult_le_pos; [|apply Rmult_le_pos]; lra'. }
exists (mk_nnR _ PSM); split => [//|/=].
apply (Rle_trans _ _ _ (point_10 M)); rewrite /M; right.
field; rewrite /R'; lra.
Qed.

End Theorem5.
